package Author::Daemon::Site::Ptr::Bond;

# Perl core
use v5.32;
use strict;
use warnings;
use experimental "signatures";

# ABSTRACT: turns baubles into trinkets
# Version
our $VERSION = '0.03';

# Perl bundled
use Carp qw(cluck longmess shortmess);

# Private local

# Private remote

# Community

1;
